import unittest

from sortwiz.base_sort import BaseSort, SortResult

class TestBaseSort(unittest.TestCase):
    sut = BaseSort()

    def test_validate(self):
        # Arrange
        items = "any str"

        # Act & Assert
        with self.assertRaises(TypeError):
            self.sut.validate(items)

    def test_sort_not_implemented(self):
         # Act & Assert
        with self.assertRaises(NotImplementedError):
            self.sut.sort([], True, 1)

        with self.assertRaises(NotImplementedError):
            self.sut.sort_asc([], True)

        with self.assertRaises(NotImplementedError):
            self.sut.sort_asc([], True)

    def test_swap(self):
        # Arrange
        items = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
        item_to_be_swapped = 998
        item_to_be_swapped_index = 0
        another_item_to_be_swapped = 999
        another_item_to_be_swapped_index = 3

        items[item_to_be_swapped_index] = item_to_be_swapped
        items[another_item_to_be_swapped_index] = another_item_to_be_swapped

        sort_result = SortResult([], 0, 0, 0)

        # Act
        BaseSort.swap(items, item_to_be_swapped_index, another_item_to_be_swapped_index, sort_result)

        # Assert
        self.assertEqual(item_to_be_swapped, items[another_item_to_be_swapped_index])
        self.assertEqual(another_item_to_be_swapped, items[item_to_be_swapped_index])


    


from datetime import datetime

from .base_sort import BaseSort, SortResult
from .errors import InvalidOrderError

class BubbleSort(BaseSort):
    @classmethod
    def sort_asc(self, items: list, in_place: bool = False) -> SortResult:
        f'''
        Applies a bubble sort in ascending order to the given list.

        Parameters
        ----------
            `items` - A list with items to sort.
            `in_place` - Whether to apply the sort into the given list or not. If True, modifies the given list, if False, not.

        Returns
        -------
            The `SortResult` with the list of items in ascending order.
        '''

        return self.sort(items, in_place, False)
    
    @classmethod
    def sort_desc(self, items: list, in_place: bool = False) -> SortResult:
        '''
        Applies a bubble sort in descending order to the given list.

        Parameters
        ----------
            `items` - A list with items to sort.
            `in_place` - Whether to apply the sort into the given list or not. If True, modifies the given list, if False, not.

        Returns
        -------
            The `SortResult` with the list of items in descending order.
        '''

        return self.sort(items, in_place, True)
    
    @classmethod
    def sort(self, items: list, in_place: bool = False, reverse: bool = False) -> SortResult:
        '''
        Applies a bubble sort to the given list.

        Parameters
        ----------
            `items` - A list with items to sort.
            `in_place` - Whether to apply the sort into the given list or not. If True, modifies the given list, if False, not.
            `reverse` - `True` for descending, `False` for ascending.

        Returns
        -------
            The `SortResult` with the list of items.
        '''
        self.validate(items)

        start_time = datetime.now()

        swapped = True
        sorted_list = items if in_place else list(items)
        sort_result = SortResult(sorted_list, 0, 0, 0)

        while swapped:
            swapped = False

            for i in range(len(sorted_list) - 1):
                if self.__apply_order_rule(sorted_list, i, reverse):
                    self.swap(sorted_list, i, i + 1, sort_result)
                    swapped = True

        end_time = datetime.now()

        sort_result.time = end_time - start_time

        return sort_result
    
    @classmethod
    def __apply_order_rule(self, items: list, i: int, reverse: bool):
        if reverse:
            return items[i] < items[i + 1]
        
        return items[i] > items[i + 1]
        



from .base_sort import SortResult
from .bubble_sort import BubbleSort

class SortWizard:
    _bubble_sort = BubbleSort()

    @staticmethod
    def bubble_sort(items: list, in_place : bool = False, reverse: bool = False) -> SortResult:
        '''
        Applies a bubble sort to the given list.

        Parameters
        ----------
            `items` - A list with items to sort.
            `in_place` - Whether to apply the sort into the given list or not. If True, modifies the given list, if False, not.
            `reverse` - `True` for descending, `False` for ascending.

        Returns
        -------
            The `SortResult` with the list of items.
        '''

        return SortWizard._bubble_sort.sort(items, in_place, reverse)
    


import unittest

from sortwiz.errors import InvalidOrderError
from sortwiz.bubble_sort import BubbleSort
from sortwiz.base_sort import SortResult

class TestBubbleSort(unittest.TestCase):
    sut = BubbleSort()

    def test_items_type(self):
        # Arrange
        items = "any str"

        # Act & Assert
        with self.assertRaises(TypeError):
            self.sut.sort(items, True)

        with self.assertRaises(TypeError):
            self.sut.sort_asc(items, True)

        with self.assertRaises(TypeError):
            self.sut.sort_asc(items, True)

    def test_in_place_True_returns_same_list_reference(self):
        # Arrange
        in_place = True

        items = [5, 8, 2, 16, 0, -5]
        items_asc = list(items)
        items_desc = list(items)

        # Act
        result = self.sut.sort(items, in_place)
        result_asc = self.sut.sort_asc(items_asc, in_place)
        result_desc = self.sut.sort_desc(items_desc, in_place)

        # Assert
        self.assertEqual(items, result.value)
        self.assertEqual(items_asc, result_asc.value)
        self.assertEqual(items_desc, result_desc.value)

    def test_in_place_False_returns_different_list_reference(self):
        # Arrange
        in_place = False

        items = [5, 8, 2, 16, 0, -5]
        items_asc = list(items)
        items_desc = list(items)

        # Act
        result = self.sut.sort(items, in_place)
        result_asc = self.sut.sort_asc(items, in_place)
        result_desc = self.sut.sort_desc(items, in_place)

        # Assert
        self.assertNotEqual(items, result.value)
        self.assertNotEqual(items_asc, result_asc.value)
        self.assertNotEqual(items_desc, result_desc.value)
    
    def test_items_is_empty_returns_zero_for_swaps_and_writes(self):
        # Arrange
        empty_items = []

        # Act
        result = self.sut.sort(empty_items, True)

        # Assert
        self.assertEqual(0, result.swaps)
        self.assertEqual(0, result.writes)
        self.assertEqual(empty_items, result.value)

    def test_items_sort_asc(self):
        # Arrange
        items = [8, 5, 6, 1, 0, -10, 6]

        expected = list(items)
        expected.sort()

        # Act
        result = self.sut.sort(items, False)
        result_asc = self.sut.sort_asc(items, False)

        # Arrange
        self.assertListEqual(expected, result.value)
        self.assertListEqual(expected, result_asc.value)

    def test_items_sort_desc(self):
        # Arrange
        items = [8, 5, 6, 1, 0, -10, 6]

        expected = list(items)
        expected.sort(reverse=True)

        # Act
        result = self.sut.sort(items, False, True)
        result_desc = self.sut.sort_desc(items, False)

        # Arrange
        self.assertListEqual(expected, result.value)
        self.assertListEqual(expected, result_desc.value)

    def test_items_swaps_asc(self):
        # Arrange
        items = [4, 3, 1, 2]

        expected = 5

        # Act
        result = self.sut.sort(items, False)
        result_asc = self.sut.sort_asc(items, False)

        # Arrange
        self.assertEqual(expected, result.swaps)
        self.assertEqual(expected, result_asc.swaps)

    def test_items_swaps_desc(self):
        # Arrange
        items = [4, 3, 1, 2]

        expected = 1

        # Act
        result = self.sut.sort(items, False, True)
        result_desc = self.sut.sort_desc(items, False)

        # Arrange
        self.assertEqual(expected, result.swaps)
        self.assertEqual(expected, result_desc.swaps)

    def test_items_writes_asc(self):
        # Arrange
        items = [4, 3, 1, 2]

        expected = 10

        # Act
        result = self.sut.sort(items, False)
        result_asc = self.sut.sort_asc(items, False)

        # Arrange
        self.assertEqual(expected, result.writes)
        self.assertEqual(expected, result_asc.writes)

    def test_items_writes_desc(self):
        # Arrange
        items = [4, 3, 1, 2]

        expected = 2

        # Act
        result = self.sut.sort(items, False, True)
        result_desc = self.sut.sort_desc(items, False)

        # Arrange
        self.assertEqual(expected, result.writes)
        self.assertEqual(expected, result_desc.writes)
    
    def test_result_type(self):
        # Act
        result = self.sut.sort([], False)
        result_asc = self.sut.sort_asc([], False)
        result_desc = self.sut.sort_desc([], False)

        # Assert
        self.assertIsInstance(result, SortResult)
        self.assertIsInstance(result_asc, SortResult)
        self.assertIsInstance(result_desc, SortResult)



class InvalidOrderError(Exception):
    def __init__(self, order) -> None:
        self.order = order
        self.message = f"Invalid sorting order given: {self.order}.\nUse SortingWizard.ASCENDING and SortingWizard.DESCENDING for valid orders."
        super().__init__(self.message)
        


from datetime import timedelta

class SortResult:
    def __init__(self, value: list, swaps: int, writes: int, time: timedelta):
        self.value = value
        self.swaps = swaps
        self.writes = writes
        self.time = time

    def __str__(self) -> str:
        return f'''
Sort result
| value (first 10) - {self.value[:10]}
| swaps - {self.swaps}
| writes - {self.writes}
| time - {self.time}
'''

class BaseSort:
    @classmethod
    def sort_asc(self, items: list, in_place: bool = False) -> SortResult:
        raise NotImplementedError()
    
    @classmethod
    def sort_desc(self, items: list, in_place: bool = False) -> SortResult:
        raise NotImplementedError()
    
    @classmethod
    def sort(self, items: list, in_place: bool = False, reverse: bool = False) -> SortResult:
        raise NotImplementedError()
    
    @staticmethod
    def validate(items) -> None:
        '''
        Validates if `items` is of type list.

        Returns
        -------
            `None`.

        Raises
        ------
            `TypeError` if items isn't a list.
        '''
        
        items_type = type(items)
        if items_type != list:
            raise TypeError(f"Parameter 'items' expected be a list. Given: {items_type}.")
    
    @staticmethod
    def swap(items: list, i: int, j: int, sorting_result: SortResult) -> None:
        items[i], items[j] = items[j], items[i]
        sorting_result.writes += 2
        sorting_result.swaps += 1
